from __future__ import division
import matplotlib.pyplot as plt
from scipy.signal import savgol_filter
from scipy.optimize import leastsq, curve_fit, root
import numpy as np
import pickle

# Pauli matrices
sx = np.array([[0, 1], [1, 0]], complex)
sy = np.array([[0 , -1j], [1j, 0]])
sz = np.array([[1, 0], [0, -1]], complex)
s0 = np.array([[1, 0], [0, 1]], complex)


class SimpleNamespace(object):
    """A simple container for parameters."""
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


def charge_operator(M):
    """
    Returns the charge operator N as a dense 2M x 2M NumPy array.
    The charge operator is represented in the phase basis, see
    Supplementary Material.
    
    Parameters:
    -----------
    M: int, odd
        Truncation of the Hilbert space.
    """
    ks = np.arange(-(M-1)/2, (M+1)/2, 1)
    ch = np.zeros((M, M)) + 0j
    for dk in range(1, M):
        x = np.pi * dk / M
        ch += - 1j * 0.5 * (((-1)**dk / np.sin(x))
                            * (np.eye(M, k=dk) - np.eye(M, k=-dk)) + 0j)
    return np.kron(np.eye(2), ch)


def Josephson_potential_operator(p, M):
    """
    Returns the potential energy of the model
    as a dense 2M x 2M NumPy array.
    
    Parameters:
    -----------
    p: SimpleNamespace instance.
        It should contain the following parameters of the model:
        `T` : float, in [0,1); transmission of the weak link.
        `Ej`: float; Josephson energy of the tunnel junction.
        `Ea`: float; Josephson energy of the weak link.
        `flux`: float; flux through the SQUID.
    M: int, odd.
        Truncation of the Hilbert space.
    """
    ks = np.arange(-(M-1)/2, (M+1)/2, 1)
    r = np.sqrt(1 - p.T)

    # The potential is a sum of three terms.
    # The first term is for the tunnel junction.
    # Vy and Vz are given in Eqs. (S7), (S8).
    V0 = lambda x: 1 - np.cos(x)
    Vy = lambda x: r * np.sin(x/2) * np.cos(r*x/2) - np.cos(x/2) * np.sin(r*x/2)
    Vz = lambda x: np.cos(x/2) * np.cos(r*x/2) + r * np.sin(x/2) * np.sin(r*x/2)
    
    V = p.Ej * np.kron(np.eye(2),
                       np.diag(np.array([V0(2 * np.pi * k / M)
                                         for k in ks])) + 0j)
    V += p.Ea * np.kron(sy,
                        np.diag(np.array([Vy(p.flux - 2 * np.pi * k / M)
                                          for k in ks])))
    V += p.Ea * np.kron(sz,
                        np.diag(np.array([Vz(p.flux - 2 * np.pi * k / M)
                                          for k in ks])))
    return V


def current_operator(p, M):
    """
    Returns the current operator of the model
    as a dense 2M x 2M NumPy array.
    
    Parameters:
    -----------
    p: SimpleNamespace instance.
        It should contain the following parameters of the model:
        `T` : float, in the interval [0,1); transmission of the weak link.
        `Ej`: float; Josephson energy of the tunnel junction.
        `Ea`: float; Josephson energy of the weak link.
        `flux`: float; flux through the SQUID.
    M: int, odd.
        Truncation of the Hilbert space.
    """
    ks = np.arange(-(M-1)/2, (M+1)/2, 1)
    r = np.sqrt(1 - p.T)

    # See Eq. (S15) for the definition of the current operator.
    I0 = lambda x: np.sin(x)
    Iy = lambda x: 0.5 * p.T * np.sin(x/2) * np.sin(r*x/2)
    Iz = lambda x: - 0.5 * p.T * np.sin(x/2) * np.cos(r*x/2)

    I = p.Ej * np.kron(np.eye(2),
                       np.diag(np.array([I0(2 * np.pi * k / M)
                                         for k in ks])) + 0j)
    I += p.Ea * np.kron(sy,
                        np.diag(np.array([Iy(p.flux - 2 * np.pi * k / M)
                                          for k in ks])))
    I += p.Ea * np.kron(sz, np.diag(np.array([Iz(p.flux - 2 * np.pi * k / M)
                                              for k in ks])))
    return I


def create_hamiltonian(p, M):
    """
    Returns the Hamiltonian of the model
    as a dense 2M x 2M NumPy array.
    
    Parameters:
    -----------
    p: SimpleNamespace instance.
        It should contain the following parameters of the model:
        `T` : float, in the interval [0,1); transmission of the weak link.
        `Ej`: float; Josephson energy of the tunnel junction.
        `Ea`: float; Josephson energy of the weak link.
        `flux`: float; flux through the SQUID.
        `Ec`: float; charging energy.
    M: int, odd.
        Truncation of the Hilbert space.
    """
    N = charge_operator(M)
    T = p.Ec * N.dot(N)
    V = Josephson_potential_operator(p, M)
    H = np.add(T, V)
    return H


def find_optimal_M(p, max_tol = 9e-8):
    """
    Returns the optimal Hilbert space truncation M
    such that low-lying eigenvalues of H are not
    affected by relevant finite-size errors.

    Parameters:
    -----------
    p: SimpleNamespace
         See docstring of create_hamiltonian()
    max_tol: float
         Required minimum error on the eigenvalues of H.

    Return:
    --------
    M: int
        Size of the Hilbert space which satisfies
        the required tolerance in the returned eigenvalues.

    Notes:
    ------
    The convergence test is very simple. The returned M is the lowest
    which satisfies the inequality |dE(M)-dE(M-4)|<max_tol, where dE(M)
    is the difference of the two lowest eigenvalyes compute for a
    Hamiltonian of size 2M x 2M.
    """
    M = 21 # Starting point of the iteration
    tol = 100
    H = create_hamiltonian(p, M)
    ens = np.linalg.eigvalsh(H)
    de = ens[1]-ens[0]
    while tol > max_tol:
        M += 4
        H = create_hamiltonian(p, M)
        ens = np.linalg.eigvalsh(H)
        de_new = ens[1]-ens[0]
        tol = np.abs(de_new - de)
        de = de_new
    return M


def current_matrix_elements(p, M):
    """
    Returns the array of current matrix elements I_n
    
    I_n = < 0 | I | n>
    
    between the n-th eigenstate of H and the ground state.
    
    Parameters:
    -----------
    p: SimpleNamespace
         See docstring of create_hamiltonian()
    M: int, odd.
        Truncation of the Hilbert space.
    """
    matrix_elements = []
    ___, evec = np.linalg.eigh(create_hamiltonian(p, M))
    wavefunctions = evec.T
    iop = current_operator(p, M)
    i_n = np.array([np.conj(wavefunctions[0]).T.dot(iop).dot(wf)
                    for wf in wavefunctions])
    return i_n


def return_wavefunction(p, M, n):
    """
    Returns the two-component wave function of the n-th eigenstate of H.
    The wave function is represented as a pair of M-dimensional NumPy arrays.
    
    Parameters:
    -----------
    p: SimpleNamespace
         See docstring of create_hamiltonian()
    M: int, odd.
        Truncation of the Hilbert space.
    n: int
        Index for the desired eigenstate.
    """
    ___, evecs = np.linalg.eigh(create_hamiltonian(p, M))
    # Picks wave function corresponding to n-th eigenvalue
    wf = evecs[:,n]
    # Defines the local (i.e., flux-dependent) basis of
    # ABS ground and excited states (g, e).
    # See Eq. (S14) in the Supplementary Material
    r = np.sqrt(1-p.T)
    phase = r * p.flux / 4
    U = np.cos(phase) * s0 - 1j * np.sin(phase) * sx
    V2 = r * np.sin(p.flux / 2) * U.dot(sy).dot(U.conj().T)
    V3 = np.cos(p.flux / 2) * U.dot(sz).dot(U.conj().T)
    ___, (gs, es) = np.linalg.eigh(V2 + V3)
    # Rotates the wave function in the (g, e) basis
    gs_wf = np.array([wf[i] * gs[0] + wf[M+i] * gs[1] for i in xrange(M)])
    es_wf = np.array([wf[i] * es[0] + wf[M+i] * es[1] for i in xrange(M)])
    return gs_wf, es_wf


# The following two functions are used for the fit to the numerical data.
# They are taken from the notebook at goo.gl/3xfGr8, with small modifications. 
def rs_model(plist, xs, ys, idx, n=2,
                 Ej=None, Ec=None, Ea=None, max_tol=9e-8, sigmas=None):
    """ Returns the list of residuals between the data points
    and the transition energies, given a list of Hamiltonian parameters.

    Parameters:
    -----------
    plist : list of floats
        List of Hamiltonian parameters (excluding flux).
    xs : 1d array
        X values of the data.
    ys : 1d array
        Y values of the data.
    sigmas: 1d array
        Y data errors, if available.
    M : int
        Size of Hamiltonian. Must be odd.
    idx : list
        Index which marks the beginning of data corresponding
        to the ABS transition line. It is used to associate each
        data point to the correct transition energy.
    n : int
        Tells which transition 0 --> n to associate to the
        putative ABS line in the data. For instance, n=2 will
        try to fit the 0 --> 2 transition to the ABS line.
        The integer n should be > 1.
    
    Notes:
    ------
    plist should contain the Hamiltonian parameters in the following
    order: Ec, Ej, Ea, T. To keep some parameters fixed during the fit,
    they should be passed separately. See instructions of fit_abs_model.
    
    Returns:
    ------
    1d array of weighted residuals.
    """
    p = SimpleNamespace()
    p.flux = np.pi # Initalize value of flux to compute optimal matrix size
    
    if not Ej and not Ec and not Ea:
        p.Ec, p.Ej, p.Ea, p.T = plist
    elif not Ec and not Ea:
        p.Ec, p.Ea, p.T = plist
        p.Ej = Ej
    elif not Ea:
        p.Ea, p.T = plist
        p.Ej, p.Ec = Ej, Ec
    else:
        p.T = plist[0]
        p.Ej, p.Ec, p.Ea = Ej, Ec, Ea
    
    # if leastsq attempts T > 1, a huge residual is returned
    if int(np.floor(p.T)) is not 0:
        return 1e12 * ys
    # Similarly, we constrain values of other energy scales
    if p.Ea > 200 or p.Ea < 0:
        return 1e12 * ys
    if p.Ej > 400 or p.Ej < 0:
        return 1e12 * ys
    if p.Ec > 50 or p.Ec < 0:
        return 1e12 * ys
    
    M = find_optimal_M(p, max_tol = max_tol) + 50
    hams = []
    for x in xs:
        p.flux = x
        hams.append(create_hamiltonian(p, M))
    evals = [np.sort(np.linalg.eigvalsh(h)) for h in hams]
    freqs = [(ev - ev[0]) for ev in evals]
    
    frequencies = np.hstack(([f[1] for f in freqs[:idx]],
                             [f[n] for f in freqs[idx:]]))
    weights = (1. / sigmas) if sigmas else 1.
    # Factor 0.5 is because measured lines are at a voltage 2eV = omega.
    # Hence, eV = 0.5 omega
    return (0.5 * frequencies - ys) * weights


def fit_model(lines, idx, p0, n=2, Ej=None, Ec=None, Ea=None, max_tol=9e-7):
    """
    Returns the best estimate for the parameters p0 and their
    covariance matrix.

    The obtain the covariance matrix, the jacobian around the
    solution returned by leastsq is multiplied by the residual variance

    s = sum(rs) / (n-m)

    where n is the number of data points and m the number of fitting
    parameters. This is the same as what is done in scipy.optimize.curve_fit.
    
    Parameters:
    -----------
    lines : list
        List of points to be fitted, one list for every
        different transition.
    idx: int
        Index which marks the beginning of data corresponding
        to the ABS transition line.
    p0 : list
        Initial guess for the fitting parameters.
    Ej : float
        Josephson energy of the conventional junction.
        If Ej=None, Ej is passed via p0 and included as fitting parameter.
        Otherwise, it is kept as a fixed parameter.
    Ec : float
        Charging energy. Same instructions as Ej.
    Ea : float
        Charging energy. Same instructions as Ej.
    n : int
        Tells which transition 0 --> n to associate to the
        putative ABS line in the data. For instance, n=2 will
        try to fit the 0 --> 2 transition to the ABS line.
        The integer n should be > 1.
    """
    fit_result = {'initial_guess' : p0}
    xs, ys= np.asarray(zip(*lines))
    
    (popt, pcov, infodict, errmsg, ier) = \
        leastsq(rs_model, x0=p0,
                args=(xs, ys, idx, n, Ej, Ec, Ea, max_tol), full_output=1)
    
    chi_squared = (np.asarray(rs_model(popt, xs, ys, idx, n,
                                       Ej, Ec, Ea, max_tol))**2).sum()
    chi_squared /= (len(xs) - len(popt))
    pcov = pcov * chi_squared
    std_devs = np.sqrt(np.diag(pcov))
    fit_result['popt'] = popt
    fit_result['chi_squared'] = chi_squared
    fit_result['infodict'] = infodict
    fit_result['pcov'] = pcov
    fit_result['std_devs'] = std_devs
    fit_result['errmsg'] = errmsg
    fit_result['ier'] = ier
    return fit_result